/*
 * Remover.java
 *
 * Created on July 10, 2001, 11:18 AM
 *
 * See LICENSE file for license conditions.
 */

package residue;

import java.io.*;
import java.util.*;
import residue.events.*;

import org.apache.bcel.classfile.*;
import org.apache.bcel.generic.*;

/**
 *
 * @author  chowells
 * @version 
 */
public class Remover extends AbstractInstrumenter
{
	public static void main(String [] args) throws IOException
	{
		Remover r = new Remover();
		r.addInstrumentationEventListener(new InstrumentationEventListener()
		{
			public void instrumentationStarted(InstrumentationEvent e)
			{
				 System.out.println("Starting removal of probes");
			} // instrumentationStarted
			
			public void instrumentationFinished(InstrumentationEvent e)
			{
				System.out.println("Finished removal of probes");
			} // instrumentationFinished
			
			public void classStarted(InstrumentationEvent e)
			{
				System.out.println("Removing probes from " + e.getClassName());
			} // classStarted
		}); // anonymous InstrumentationEventListener
		
		Set toRemove = new HashSet();
		
		for (int i = 0; i < args.length; i++)
			toRemove.add(new File(args[i]));
		
		r.remove(toRemove);
	} // main
	
	/**
	 * The argument is the set of files (represented as files) to
	 * remove instrumentation from. This process removes all
	 * function calls added, but leaves the entries added to the
	 * constant pool.  This is sufficient to eliminate the need
	 * for the runtime classes during execution of the previously
	 * instrumented classes.
	 * 
	 * Because removed probes may or may not belong to the current
	 * instrumentation configuration, the current config is removed.
	 */
	public void remove(Set s) throws IOException
	{
		int size = s.size();
		instrumentationStarted(null, size);
		
		Iterator it = s.iterator();
		while (it.hasNext())
		{
			File oldf = (File)it.next();
			classStarted(oldf.toString(), size);
			
			File f = oldf.getCanonicalFile();
			
			removeInstrumentation(f);
			
			String name = f.getAbsolutePath();
			File g = new File(name.substring(0, name.length() - 6).concat(".gretel"));
			g.delete();
			
			size--;
		} // while
		instrumentationFinished(null, 0);
	} // remove
	
	/**
	 * Method that actually removes instrumentation.
	 */
	private void removeInstrumentation(File f) throws IOException
	{
		if (!f.toString().endsWith(".class"))
		{
			System.out.println("Unrecognized file type " + f);
			return;
		}
		
		JavaClass jc = new ClassParser(f.toString()).parse();
		
		ConstantPoolGen cp = new ConstantPoolGen(jc.getConstantPool());
		
		int hit = cp.lookupMethodref("residue.runtime.Monitor", "hit", "(I)V");
		int init = cp.lookupMethodref("residue.runtime.Monitor", "init", "(Ljava/lang/String;)V");
		
		// remove calls to either of the above methods
		Method [] methods = jc.getMethods();
		for (int i = 0; i < methods.length; i++)
		{
			if (methods[i].isAbstract() || methods[i].isNative() ||
				methods[i].getCode() == null) continue;
			
			MethodGen mg = new MethodGen(methods[i], jc.getClassName(), cp);
			InstructionList il = mg.getInstructionList();
			
			InstructionHandle curr = il.getStart();
			while (curr != null)
			{
				Instruction in = curr.getInstruction();
				
				if (in instanceof INVOKESTATIC)
				{
					INVOKESTATIC is = (INVOKESTATIC)in;
					
					if (is.getIndex() == hit || is.getIndex() == init)
					{
						InstructionHandle start = curr.getPrev();
						InstructionHandle end = curr;
						
						curr = curr.getNext();
						
						InstructionTargeter [] its = start.getTargeters();
						
						if (its != null)
							for (int j = 0; j < its.length; j++)
								its[j].updateTarget(start, curr);
						
						try
						{
							il.delete(start, end);
						}
						catch (TargetLostException e)
						{
							System.out.println("Target lost unexpectedly removing call in Remover.");
							System.out.println("Please check code");
						}
					}
					else
					{
						curr = curr.getNext();
					}
				}
				else
				{
					curr = curr.getNext();
				}
				
			} // while
			
			mg.setMaxStack();
			methods[i] = mg.getMethod();
			il.dispose();
		} // for
		
		jc.setConstantPool(cp.getFinalConstantPool());
		
		jc.dump(f.toString());
	} // removeInstrumentation
	
} // class Remover
